#   Conversion to MCNP v0.0
#   Only one solid and planar surfaces
#

import logging
import math
from collections import OrderedDict

import FreeCAD
import Part

from ..conversion import cell_definition as CD
from ..utils import functions as UF
from ..utils import geometry_gu as GU
from ..utils.basic_functions_part1 import (
    is_in_line,
    is_in_plane,
    is_parallel,
    is_same_value,
)
from ..utils.basic_functions_part2 import is_duplicate_in_list

logger = logging.getLogger("general_logger")

twoPi = math.pi * 2


def split_full_cylinder(solid, options, tolerances, numeric_format):
    explode = []
    bases = [solid]
    while True:
        new_bases = []
        for base in bases:
            cut_solids = cut_full_cylinder(base, options, tolerances, numeric_format)
            if len(cut_solids) == 1:
                explode.extend(cut_solids)
            else:
                new_bases.extend(cut_solids)
        if len(new_bases) == 0:
            break
        else:
            bases = new_bases

    return Part.makeCompound(explode)


def cut_full_cylinder(solid, options, tolerances, numeric_format):
    solid_gu = GU.SolidGu(solid, tolerances=tolerances)
    surfaces = UF.SurfacesDict()
    flag_inv = CD.is_inverted(solid_gu.solid)
    universe_box = solid.BoundBox

    for face in solid_gu.Faces:
        surf = str(face.Surface)
        if surf == "<Cylinder object>":
            if flag_inv:
                orient = "Reversed" if face.Orientation == "Forward" else "Forward"
            else:
                orient = face.Orientation

            u1, u2, v1, v2 = face.ParameterRange
            angle = abs(u2 - u1)

            # closed convex cylinder
            if abs(angle % twoPi) < 1e-2 and orient == "Forward":
                dir = face.Surface.Axis
                orig = face.Surface.Center
                rad = face.Surface.Radius
                dim_l = face.ParameterRange[3] - face.ParameterRange[2]
                cylinder = UF.GeounedSurface(("Cylinder", (orig, dir, rad, dim_l)), universe_box)
                cylinder.build_surface()
                surfaces.add_cylinder(cylinder, options, tolerances, numeric_format, False)

                # add planes if cylinder axis is cut by a plane (plane quasi perpedicular to axis)
                for p in cyl_bound_planes(solid_gu, face, universe_box):
                    p.build_surface()
                    surfaces.add_plane(p, options, tolerances, numeric_format, False)
                break

    planes = []
    for P in ("PX", "PY", "PZ", "P"):
        planes.extend(surfaces[P])
    planes = sort_planes(planes, True)

    if len(planes) == 0:
        return [solid]
    if len(planes[-1]) < options.nPlaneReverse:
        planes.reverse()

    cut = False
    for pp in planes:
        # cut with more external parallel planes
        pp[0].build_surface()
        if len(pp) != 1:
            pp[-1].build_surface()
            tools = (pp[0].shape, pp[-1].shape)
        else:
            tools = (pp[0].shape,)

        try:
            comsolid = UF.split_bop(solid, tools, options.splitTolerance, options)
        except:
            comsolid = solid
        if len(comsolid.Solids) > 1:
            out_solid = comsolid.Solids
            cut = True
            break

    if cut:
        return out_solid

    tool = (surfaces["Cyl"][0].shape,)
    try:
        comsolid = UF.split_bop(solid, tool, options.splitTolerance, options)
    except:
        comsolid = solid

    if len(comsolid.Solids) > 1:
        out_solid = comsolid.Solids
    else:
        out_solid = [solid]

    return out_solid


def gen_plane(pos, normal, diag):
    plane = Part.makePlane(diag, diag, pos, normal)
    vec_on_plane = plane.Vertexes[3].Point.sub(plane.Vertexes[0].Point)
    new_pos = plane.Vertexes[0].Point.sub(vec_on_plane)
    plane_center = Part.makePlane(2.0 * diag, 2.0 * diag, new_pos, normal)
    return plane_center


def other_face_edge(current_edge, current_face, solid):
    for face in solid.Faces:
        if face.isSame(current_face):
            continue
        for edge in face.Edges:
            if current_edge.isSame(edge):
                return face


def cyl_bound_planes(solid, face, boundBox):
    Edges = face.OuterWire.Edges
    planes = []
    for e in Edges:
        try:
            curve = str(e.Curve)
        except:
            curve = "none"

        adjacent_face = other_face_edge(e, face, solid)
        if adjacent_face is not None:
            if type(adjacent_face.Surface) is GU.TorusGu:
                continue  # doesn't create plane if other face is a torus
            if face.Surface.isSameSurface(adjacent_face.Surface):
                continue  # doesn't create plane if other face has same surface

        if curve[0:6] == "Circle":
            dir = e.Curve.Axis
            center = e.Curve.Center
            dim1 = e.Curve.Radius
            dim2 = e.Curve.Radius
            plane = UF.GeounedSurface(("Plane", (center, dir, dim1, dim2)), boundBox)
            planes.append(plane)

        elif curve == "<Ellipse object>":
            dir = e.Curve.Axis
            center = e.Curve.Center
            dim1 = e.Curve.MinorRadius
            dim2 = e.Curve.MajorRadius
            plane = UF.GeounedSurface(("Plane", (center, dir, dim1, dim2)), boundBox)
            planes.append(plane)

    return planes


def torus_bound_planes(solid, face, boundBox, tolerances):
    params = face.ParameterRange
    planes = []
    if is_same_value(params[1] - params[0], twoPi, tolerances.value):
        return planes

    Edges = face.OuterWire.Edges

    for e in Edges:
        try:
            curve = str(e.Curve)
        except:
            curve = "none"

        adjacent_face = other_face_edge(e, face, solid)
        if adjacent_face is not None:
            if face.Surface.isSameSurface(adjacent_face.Surface):
                continue  # doesn't create plane if other face has same surface

        if curve[0:6] == "Circle":
            dir = e.Curve.Axis
            if not is_parallel(dir, face.Surface.Axis, tolerances.angle):
                center = e.Curve.Center
                dim1 = e.Curve.Radius
                dim2 = e.Curve.Radius
                plane = UF.GeounedSurface(("Plane", (center, dir, dim1, dim2)), boundBox)
                planes.append(plane)

        elif curve == "<Ellipse object>":
            dir = e.Curve.Axis
            center = e.Curve.Center
            dim1 = e.Curve.MinorRadius
            dim2 = e.Curve.MajorRadius
            plane = UF.GeounedSurface(("Plane", (center, dir, dim1, dim2)), boundBox)
            planes.append(plane)

        elif curve == "<BSplineCurve object>":
            planeParams = plane_spline_curve(e, tolerances)
            if planeParams is not None:
                plane = UF.GeounedSurface(("Plane", planeParams), boundBox)
                planes.append(plane)

    return planes


def plane_spline_curve(edge, tolerances):

    normal = edge.derivative1At(0).cross(edge.derivative1At(0.5))
    if normal.Length == 0:
        return None
    normal.normalize()
    curve_2d = True
    for p in (0.25, 0.75, 1):
        # check if derivative orthogonal to curve normal vector
        if abs(normal.dot(edge.derivative1At(p))) > tolerances.value:
            curve_2d = False
            break

    r = edge.valueAt(0.25) - edge.valueAt(0.75)
    if curve_2d:
        return (edge.valueAt(0), normal, r.Length, r.Length)
    else:
        return None


def extract_surfaces(solid, kind, universe_box, options, tolerances, numeric_format, MakeObj):
    if kind == "All":
        fuzzy = True
        solid_parts = []
        for s in solid.Solids:
            solid_parts.append(GU.SolidGu(s, tolerances=tolerances))
    else:
        fuzzy = False
        if kind == "Plane3Pts":
            P3P = True
        else:
            P3P = False
        solid_parts = [GU.SolidGu(solid, tolerances=tolerances, plane3Pts=P3P)]

    surfaces = UF.SurfacesDict()

    for solid_GU in solid_parts:

        # Get the parameter of all faces of the solid
        # Add auxillary planes for Cylinders and Cones
        for face in solid_GU.Faces:
            surf = str(face.Surface)
            if surf == "<Plane object>" and kind in ["Planes", "All"]:
                normal = face.Surface.Axis
                pos = face.CenterOfMass
                dim1 = face.ParameterRange[1] - face.ParameterRange[0]
                dim2 = face.ParameterRange[3] - face.ParameterRange[2]
                plane = UF.GeounedSurface(("Plane", (pos, normal, dim1, dim2)), universe_box)
                if MakeObj:
                    plane.build_surface()
                surfaces.add_plane(plane, options, tolerances, numeric_format, fuzzy)

            elif surf == "<Cylinder object>":
                dir = face.Surface.Axis
                orig = face.Surface.Center
                rad = face.Surface.Radius
                dim_l = face.ParameterRange[3] - face.ParameterRange[2]
                if kind in ["Cyl", "All"]:
                    cylinder = UF.GeounedSurface(("Cylinder", (orig, dir, rad, dim_l)), universe_box)
                    if MakeObj:
                        cylinder.build_surface()
                    surfaces.add_cylinder(cylinder, options, tolerances, numeric_format, fuzzy)

                if kind in ["Planes", "All"]:
                    # add planes if cylinder axis is cut by a plane (plane quasi perpedicular to axis)
                    for p in cyl_bound_planes(solid_GU, face, universe_box):
                        if MakeObj:
                            p.build_surface()
                        surfaces.add_plane(p, options, tolerances, numeric_format, False)

            elif surf == "<Cone object>":
                dir = face.Surface.Axis
                apex = face.Surface.Apex
                half_angle = face.Surface.SemiAngle
                dim_l = face.ParameterRange[3] - face.ParameterRange[2]
                dimR = face.Surface.Radius
                if kind in ["Cone", "All"]:
                    cone = UF.GeounedSurface(("Cone", (apex, dir, half_angle, dim_l, dimR)), universe_box)
                    if MakeObj:
                        cone.build_surface()
                    surfaces.add_cone(cone, tolerances)

                if kind in ["Planes", "All"]:
                    for p in cyl_bound_planes(solid_GU, face, universe_box):
                        if MakeObj:
                            p.build_surface()
                        surfaces.add_plane(p, options, tolerances, numeric_format, False)

            elif surf[0:6] == "Sphere" and kind in ["Sph", "All"]:
                rad = face.Surface.Radius
                pnt = face.Surface.Center
                sphere = UF.GeounedSurface(("Sphere", (pnt, rad)), universe_box)
                if MakeObj:
                    sphere.build_surface()
                surfaces.add_sphere(sphere, tolerances)

                if kind in ["Planes", "All"]:
                    for p in cyl_bound_planes(solid_GU, face, universe_box):
                        if MakeObj:
                            p.build_surface()
                        surfaces.add_plane(p, options, tolerances, numeric_format, False)

            elif surf == "<Toroid object>":
                if kind in ["Tor", "All"]:
                    radMaj = face.Surface.MajorRadius
                    radMin = face.Surface.MinorRadius
                    center = face.Surface.Center
                    dir = face.Surface.Axis
                    torus = UF.GeounedSurface(("Torus", (center, dir, radMaj, radMin)), universe_box)
                    if MakeObj:
                        torus.build_surface()
                    surfaces.add_torus(torus, tolerances)

                if kind in ["Planes", "All"]:
                    for p in torus_bound_planes(solid_GU, face, universe_box, tolerances):
                        if MakeObj:
                            p.build_surface()
                        surfaces.add_plane(p, options, tolerances, numeric_format, False)

            elif surf == "<Plane object>" and kind == "Plane3Pts":
                pos = face.CenterOfMass
                normal = face.Surface.Axis
                dim1 = face.ParameterRange[1] - face.ParameterRange[0]
                dim2 = face.ParameterRange[3] - face.ParameterRange[2]
                points = tuple(v.Point for v in face.Vertexes)

                plane = UF.GeounedSurface(("Plane3Pts", (pos, normal, dim1, dim2, points)), universe_box)
                if MakeObj:
                    plane.build_surface()
                surfaces.add_plane(plane, options, tolerances, numeric_format, fuzzy)

    return surfaces


#   Check if to faces are joint
def contiguous_face(face1, face2, tolerances):
    return face1.distToShape(face2)[0] < tolerances.distance


def same_faces(Faces, tolerances):
    Connection = OrderedDict()
    if len(Faces) == 1:
        return []

    for i, face1 in enumerate(Faces):
        Couples = []
        if not Faces[i + 1 :]:
            continue
        for j, face2 in enumerate(Faces[i + 1 :]):
            if contiguous_face(face1, face2, tolerances):

                Couples.append(i + 1 + j)

        Connection[i] = Couples

    lista = Connection[0]
    Connection.popitem(0)

    if len(Connection) == 0:  # solo los elementos de la lista de la superficie 0
        return lista

    if not lista:  # ninguna face está conecta conecta con la superficie 0
        return lista

    for elem in Connection:
        if elem in lista:  # que la key esta en lista implica que sus dependencias estan
            lista.extend(Connection[elem])
        else:
            for elem2 in Connection[elem]:
                if elem2 in lista:  # si una de sus dependencias esta en lista lo esta la clave
                    lista.append(elem)

    return list(set(lista))


# Tolerance in this function are not the general once
# function should be reviewed
def gen_plane_cylinder(face, solid, tolerances):
    Surf = face.Surface
    rad = Surf.Radius
    if face.Area < 1e-2:
        return None

    UVNodes = []
    face_index_0 = [solid.Faces.index(face)]

    try:
        face.tessellate(0.1)
        UVNodes.append(face.getUVNodes())
    except RuntimeError:
        PR = face.ParameterRange
        UVNode1 = (PR[0], PR[2])
        UVNode2 = (PR[1], PR[3])
        UVNodes.append([UVNode1, UVNode2])

    for i, face2 in enumerate(solid.Faces):

        if str(face2.Surface) == "<Cylinder object>" and not (face2.isEqual(face)):
            if (
                face2.Surface.Axis.isEqual(face.Surface.Axis, 1e-5)
                and face2.Surface.Radius == rad
                and is_in_line(face2.Surface.Center, face.Surface.Axis, face.Surface.Center)
            ):
                face_index_0.append(i)

    # prueba same_faces, parece ok
    Faces_p = []
    for ind in face_index_0:
        Faces_p.append(solid.Faces[ind])

    face_index = [face_index_0[0]]  # la face de entrada

    for k in same_faces(Faces_p, tolerances):
        face_index.append(face_index_0[k])

    # commented to let plane cut rounded corner
    # if len(face_index_0)==len(face_index): #Esto evita un corte creo!!! no debería ser así en conversion
    #    return None

    for ind in reversed(face_index[1:]):
        if solid.Faces[ind].Area <= 1e-3:
            face_index.remove(ind)

        else:
            face2 = solid.Faces[ind]
            try:
                face2.tessellate(0.1)
                UVNodes.append(face2.getUVNodes())
            except RuntimeError:
                PR = face2.ParameterRange
                UVNode1 = (PR[0], PR[2])
                UVNode2 = (PR[1], PR[3])
                UVNodes.append([UVNode1, UVNode2])

    AngleRange = 0.0

    Uval = []
    for index in face_index:
        Range = solid.Faces[index].ParameterRange
        AngleRange = AngleRange + abs(Range[1] - Range[0])
        # if not(Range[0] in Uval) and not(Range[1] in Uval):
        Uval.append(Range[0])
        Uval.append(Range[1])

    if twoPi - AngleRange < 1.0e-2 or AngleRange < 1.0e-2:
        return None

    Uval_str_cl = []

    for i, elem1 in enumerate(Uval):
        num_str1 = f"{elem1:11.4E}"
        if abs(elem1) < 1.0e-5:
            num_str1 = "%11.4E" % 0.0

        if not (is_duplicate_in_list(num_str1, i, Uval)):
            Uval_str_cl.append(num_str1)

    if len(Uval_str_cl) < 2:
        logger.info("gen_plane_cylinder : Uval_str_cl should no be void")
        return None

    face_index_2 = [face_index[0], face_index[0]]

    Node_min = UVNodes[0][0]
    Node_max = UVNodes[0][1]

    dif1_0 = abs(float(Uval_str_cl[0]) - Node_min[0])
    dif2_0 = abs(float(Uval_str_cl[1]) - Node_max[0])

    # searching for minimum and maximum angle points

    for j, Nodes in enumerate(UVNodes):
        for elem in Nodes:
            dif1 = abs(float(Uval_str_cl[0]) - elem[0])
            dif2 = abs(float(Uval_str_cl[1]) - elem[0])

            if dif1 < dif1_0:
                Node_min = elem
                face_index_2[0] = face_index[j]
                dif1_0 = dif1
            if dif2 < dif2_0:
                Node_max = elem
                face_index_2[1] = face_index[j]
                dif2_0 = dif2

    V1 = solid.Faces[face_index_2[0]].valueAt(Node_min[0], Node_min[1])
    V2 = solid.Faces[face_index_2[1]].valueAt(Node_max[0], Node_max[1])

    if V1.isEqual(V2, 1e-5):
        logger.error("in the additional plane definition")
        return None

    normal = V2.sub(V1).cross(face.Surface.Axis)

    plane = Part.Plane(V1, normal).toShape()

    return plane


# Tolerance in this function are not the general once
# function should be reviewed
def gen_plane_cone(face, solid, tolerances):

    if face.Area < 1e-2:
        return None

    UVNodes = []
    face_index_0 = [solid.Faces.index(face)]

    try:
        face.tessellate(0.1)
        UVNodes.append(face.getUVNodes())
    except RuntimeError:
        PR = face.ParameterRange
        UVNode1 = (PR[0], PR[2])
        UVNode2 = (PR[1], PR[3])
        UVNodes.append([UVNode1, UVNode2])

    for i, face2 in enumerate(solid.Faces):

        if str(face2.Surface) == "<Cone object>" and not (face2.isEqual(face)):

            if (
                face2.Surface.Axis.isEqual(face.Surface.Axis, 1e-5)
                and face2.Surface.Apex.isEqual(face.Surface.Apex, 1e-5)
                and (face2.Surface.SemiAngle - face.Surface.SemiAngle) < 1e-6
            ):

                face_index_0.append(i)

    Faces_p = []
    for ind in face_index_0:
        Faces_p.append(solid.Faces[ind])

    face_index = [face_index_0[0]]  # la face de entrada

    for k in same_faces(Faces_p, tolerances):
        face_index.append(face_index_0[k])

    # same as cylinder commennt
    # if len(face_index_0)==len(face_index):
    #    return None

    for ind in reversed(face_index[1:]):
        if solid.Faces[ind].Area <= 1e-3:
            face_index.remove(ind)
        else:
            face2 = solid.Faces[ind]
            try:
                face2.tessellate(0.1)
                UVNodes.append(face2.getUVNodes())
            except RuntimeError:
                PR = face2.ParameterRange
                UVNode1 = (PR[0], PR[2])
                UVNode2 = (PR[1], PR[3])
                UVNodes.append([UVNode1, UVNode2])

    AngleRange = 0.0

    Uval = []

    for index in face_index:
        Range = solid.Faces[index].ParameterRange
        AngleRange = AngleRange + abs(Range[1] - Range[0])
        Uval.append(Range[0])
        Uval.append(Range[1])
    if twoPi - AngleRange < 1.0e-2 or AngleRange < 1e-2:
        return None

    Uval_str_cl = []

    for i, elem1 in enumerate(Uval):
        num_str1 = f"{elem1:11.4E}"
        if abs(elem1) < 1.0e-5:
            num_str1 = "%11.4E" % 0.0
        if not (is_duplicate_in_list(num_str1, i, Uval)):
            Uval_str_cl.append(num_str1)

    if len(Uval_str_cl) < 2:
        logger.info("gen_plane_cone : Uval_str_cl should no be void")
        return None

    face_index_2 = [face_index[0], face_index[0]]

    Node_min = UVNodes[0][0]
    Node_max = UVNodes[0][1]
    dif1_0 = abs(float(Uval_str_cl[0]) - Node_min[0])
    dif2_0 = abs(float(Uval_str_cl[1]) - Node_max[0])

    # searching for minimum and maximum angle points
    for j, Nodes in enumerate(UVNodes):
        for elem in Nodes:
            dif1 = abs(float(Uval_str_cl[0]) - elem[0])
            dif2 = abs(float(Uval_str_cl[1]) - elem[0])

            if dif1 < dif1_0:
                Node_min = elem
                face_index_2[0] = face_index[j]
                dif1_0 = dif1
            if dif2 < dif2_0:
                Node_max = elem
                face_index_2[1] = face_index[j]
                dif2_0 = dif2

    V1 = solid.Faces[face_index_2[0]].valueAt(Node_min[0], Node_min[1])
    V2 = solid.Faces[face_index_2[1]].valueAt(Node_max[0], Node_max[1])

    if V1.isEqual(V2, 1e-5):
        logger.error("in the additional plane definition")
        return None

    # normal=V2.sub(V1).cross(face.Surface.Axis)

    # plane=Part.Plane(V1,normal).toShape()
    plane = Part.Plane(V1, V2, face.Surface.Apex).toShape()

    return plane


def plane_2nd_order(solid_GU, face, flag_inv, tolerances, convex=True):
    planes = []

    if face is None:
        for face in solid_GU.Faces:
            if flag_inv:
                orient_temp = face.Orientation
                if orient_temp == "Forward":
                    orient = "Reversed"
                elif orient_temp == "Reversed":
                    orient = "Forward"
            else:
                orient = face.Orientation

            if convex and orient == "Reversed":  # convex face condition
                pp = None
                if str(face.Surface) == "<Cylinder object>":
                    pp = gen_plane_cylinder(face, solid_GU, tolerances)
                elif str(face.Surface) == "<Cone object>":
                    pp = gen_plane_cone(face, solid_GU, tolerances)
                if pp is not None:
                    planes.append(pp)
            # elif str(face.Surface)[0:6] == 'Sphere':
            #    return gen_plane_sphere(face,solid_GU)

    else:
        if flag_inv:
            orient_temp = face.Orientation
            if orient_temp == "Forward":
                orient = "Reversed"
            elif orient_temp == "Reversed":
                orient = "Forward"
        else:
            orient = face.Orientation

        if not convex and orient == "Forward":
            if str(face.Surface) == "<Cylinder object>":
                pp = gen_plane_cylinder(face, solid_GU, tolerances)
            elif str(face.Surface) == "<Cone object>":
                pp = gen_plane_cone(face, solid_GU, tolerances)
            if pp is not None:
                planes.append(pp)

    return planes


def split_planes(Solids, universe_box, options, tolerances, numeric_format):
    Bases = Solids[:]
    simpleSolid = []

    # Cut with real planes defining solid faces
    while True:
        newBases = []
        for base in Bases:
            cut_solids = split_p_planes_new(base, universe_box, options, tolerances, numeric_format)
            if len(cut_solids) == 1:
                simpleSolid.extend(cut_solids)
            else:
                newBases.extend(cut_solids)
        if len(newBases) == 0:
            break
        else:
            Bases = newBases

    return simpleSolid, 0


class ParallelPlanes:
    def __init__(self, plane):
        self.Axis = plane.Surf.Axis
        self.elements = [plane]
        self.count = 1

    def append(self, plane):
        if is_parallel(plane.Surf.Axis, self.Axis, 0.1):
            self.elements.append(plane)
            self.count += 1
            return True
        else:
            return False

    def sort(self):
        pos = []
        for i, p in enumerate(self.elements):
            d = self.Axis.dot(p.Surf.Position)
            pos.append((d, i))
        pos.sort()

        sort = []
        for p in pos:
            sort.append(self.elements[p[1]])
        self.elements = sort


def sort_planes(PlaneList, sortElements=False):
    if not PlaneList:
        return []
    newList = [ParallelPlanes(PlaneList[0])]
    for p in PlaneList[1:]:
        found = False
        for pp in newList:
            if pp.append(p):
                found = True
                break
        if not found:
            newList.append(ParallelPlanes(p))

    lenList = []
    for i, pp in enumerate(newList):
        if sortElements:
            pp.sort()
        lenList.append((pp.count, i))
    lenList.sort()

    sortedPlanes = []
    for l in lenList:
        sortedPlanes.append(newList[l[1]].elements)

    return sortedPlanes


def split_p_planes_new(solid, universe_box, options, tolerances, numeric_format):
    SPlanes = extract_surfaces(solid, "Planes", universe_box, options, tolerances, numeric_format, False)

    Planes = []
    for P in ("PX", "PY", "PZ", "P"):
        Planes.extend(SPlanes[P])

    Planes = sort_planes(Planes)

    if len(Planes) == 0:
        return [solid]
    if len(Planes[-1]) < options.nPlaneReverse:
        Planes.reverse()
    out_solid = [solid]
    for pp in Planes:
        for p in pp:
            p.build_surface()
        tools = tuple(p.shape for p in pp)
        comsolid = UF.split_bop(solid, tools, options.splitTolerance, options)

        if len(comsolid.Solids) > 1:
            out_solid = comsolid.Solids
            break
    return out_solid


def split_2nd_order(Solids, universe_box, options, tolerances, numeric_format):
    err = 0
    Base = Solids
    for kind in ["Cyl", "Cone", "Sph", "Tor"]:
        kindBase = []
        while True:
            cutBase = []
            for solid in Base:
                Surfaces = extract_surfaces(
                    solid,
                    kind,
                    universe_box,
                    options,
                    tolerances,
                    numeric_format,
                    False,
                )
                if len(Surfaces[kind]) == 0:
                    kindBase.append(solid)
                else:
                    simple = True
                    for s in Surfaces[kind]:
                        s.build_surface()
                        try:
                            comsolid = UF.split_bop(solid, [s.shape], options.splitTolerance, options)
                            solidsInCom = []
                            for s in comsolid.Solids:
                                if s.Volume > 1e-9:
                                    solidsInCom.append(s)

                            if len(solidsInCom) > 1:
                                simple = False
                                cutBase.extend(solidsInCom)
                                break
                        except:
                            logger.info("Failed split base with {s.shape.Faces[0].Surface} surface")
                            err += 2

                    if simple:
                        kindBase.append(solid)

            if len(cutBase) == 0:
                Base = kindBase
                break
            else:
                Base = cutBase

    return Base, err


def split_2nd_order_planes(Solids, options, tolerances):
    err = 0
    simpleSolid = []
    Bases = Solids
    while True:
        newBases = []
        for base in Bases:
            cut_solids, err = split_2nd_o_plane(base, options, tolerances)
            if len(cut_solids) == 1:
                simpleSolid.extend(cut_solids)
            else:
                newBases.extend(cut_solids)
        if len(newBases) == 0:
            break
        else:
            Bases = newBases

    return simpleSolid, err


def split_2nd_o_plane(solid, options, tolerances):

    err = 0
    flag_inv = CD.is_inverted(solid)
    solid_GU = GU.SolidGu(solid, tolerances=tolerances)
    planes = plane_2nd_order(solid_GU, None, flag_inv, tolerances, convex=True)
    if not planes:
        return [solid], err

    for p in planes:
        comsolid = UF.split_bop(solid, [p], options.splitTolerance, options)
        if not comsolid.Solids:
            comsolid = solid
            continue
        if len(comsolid.Solids) > 1:
            removed, err = remove_solids(comsolid.Solids)
            comsolid = Part.makeCompound(removed)
            if len(removed) > 1:
                break

    return comsolid.Solids, err


def remove_solids(Solids):
    err = 0
    Vol_tol = 1e-2
    Solids_Clean = []
    for solid in Solids:
        if solid.Volume <= Vol_tol:
            logger.warning(f"remove_solids degenerated solids are produced {solid.Volume}")
            err = 2
            continue
        Solids_Clean.append(solid)

    return Solids_Clean, err


def split_component(solidShape, universe_box, options, tolerances, numeric_format):
    err = 0
    err2 = 0

    Volini = solidShape.Volume
    Solids = solidShape.Solids
    # Split with explicit planes bounding the solid and
    # implicit planes interface of two 2nd order surfaces
    split0, err = split_planes(Solids, universe_box, options, tolerances, numeric_format)
    # Split with explicit 2nd order surfaces bounding the solid

    split1, err1 = split_2nd_order(split0, universe_box, options, tolerances, numeric_format)
    err += err1

    split, err2 = split_2nd_order_planes(split1, options, tolerances)
    err += err2
    Pieces = []
    for part in split:
        if part.Volume <= 1e-10:
            logger.warning(
                f"split_component degenerated solids are produced {part.Volume}",
            )
            err += 2
            continue
        Pieces.append(part)

    comsolid = Part.makeCompound(Pieces)

    Volend = comsolid.Volume
    Volch = (Volini - Volend) / Volini

    if abs(Volch) > 1e-2:  # 1% of Volume change
        logger.warning(f"Volume has changed after decomposition: {Volch:11.4E}")
        err += 4

    return comsolid, err


# TODO rename this function as there are two with the name name
def SplitSolid(solidShape, universe_box, options, tolerances, numeric_format):

    solid_parts = []

    for solid in solidShape.Solids:

        explode = split_full_cylinder(solid, options, tolerances, numeric_format)
        piece, err = split_component(explode, universe_box, options, tolerances, numeric_format)
        solid_parts.append(piece)

    return Part.makeCompound(solid_parts), err



import BOPTools.SplitAPI

from .buildSolidCell import FuseSolid
from .Utils.booleanFunction import BoolSequence


def buildCAD(UnivCell, data, config):

    UniverseCut = True
    if "Ustart" not in config.keys():
        config["Ustart"] = 0
    if "levelMax" not in config.keys():
        config["levelMax"] = "all"
    UnivCell.name = 0
    UnivCell.Fill = config["Ustart"]

    # read all surfaces definition
    if config["format"] == "mcnp":
        factor = 10
    else:
        factor = 1

    modelSurfaces = data.GetSurfaces(scale=factor)  # scale change cm in mcnp to mm in CAD Obj

    # read Cells and group into universes
    print(config)
    levels, UniverseCells, modelSurfaces = data.GetFilteredCells(modelSurfaces, config)

    # assign to each cell the surfaces belonging to the cell
    AssignSurfaceToCell(UniverseCells, modelSurfaces)

    #    print(UniverseCells[0][120].definition.str)
    #    print(UniverseCells[0][120].surfaces)
    #    CT=build_c_table_from_solids(UnivCell.shape,UniverseCells[0][70],option='full')
    #    print(CT)
    #    simply = BoolSequence(UniverseCells[0][70].definition.str)
    #    print('antesSimply',simply)
    #    simply.simplify(CT)
    #    print('despues',simply)
    # exit()

    # dictionnary of the cells filled with a given Universe U
    # universeContainers = get_universe_containers(levels,UniverseCells)

    UnivCell.level = None
    levelMax = config["levelMax"]
    Ustart = config["Ustart"]
    if levelMax == "all":
        levelMax = len(levels)

    for lev, Univ in levels.items():
        if Ustart in Univ:
            UnivCell.level = lev - 1
            break
    startInfo = (Ustart, levelMax)

    return BuildUniverse(startInfo, UnivCell, UniverseCells, universeCut=UniverseCut)


def interferencia(container, cell, mode="slice"):

    if mode == "common":
        return cell.shape.common(container.shape)

    Base = cell.shape
    Tool = (container.shape,)

    solids = BOPTools.SplitAPI.slice(Base, Tool, "Split", tolerance=0).Solids
    cellParts = []
    for s in solids:
        if container.shape.isInside(s.CenterOfMass, 0.0, False):
            cellParts.append(s)

    if not cellParts:
        return cell.shape
    else:
        return FuseSolid(cellParts)


def AssignSurfaceToCell(UniverseCells, modelSurfaces):

    for Uid, uniCells in UniverseCells.items():
        for cId, c in uniCells.items():
            c.setSurfaces(modelSurfaces)


def get_universe_containers(levels, Universes):
    Ucontainer = {}
    for lev in range(1, len(levels)):
        for U, name in levels[lev]:
            UFILL = Universes[U][name].FILL
            if UFILL in Ucontainer.keys():
                Ucontainer[UFILL].append((U, name, lev))
            else:
                Ucontainer[UFILL] = [(U, name, lev)]
    return Ucontainer


def BuildUniverse(startInfo, ContainerCell, AllUniverses, universeCut=True, duplicate=False):
    CADUniverse = []

    Ustart, levelMax = startInfo
    Universe = AllUniverses[Ustart]

    print(f"Build Universe {ContainerCell.FILL} in container cell {ContainerCell.name}")
    fails = []
    for NTcell in Universe.values():
        if duplicate:
            if NTcell.shape:
                buildShape = False
                if ContainerCell.CurrentTR:
                    cell = NTcell.copy()
                    cell.transformSolid(ContainerCell.CurrentTR)
                else:
                    cell = NTcell
            else:
                CTRF = None
                buildShape = True
                if ContainerCell.CurrentTR:
                    CC = ContainerCell.copy()
                    CC.transformSolid(CC.CurrentTR, reverse=True)
                else:
                    CC = ContainerCell
        else:
            CTRF = ContainerCell.CurrentTR
            CC = ContainerCell
            buildShape = True
            NTcell = NTcell.copy()

        if buildShape:
            print(f"Level :{CC.level + 1}  build Cell {NTcell.name} ")
            if type(NTcell.definition) is not BoolSequence:
                NTcell.definition = BoolSequence(NTcell.definition.str)

            bBox = ContainerCell.shape.BoundBox

            debug = False
            if debug:
                NTcell.buildShape(bBox, surfTR=CTRF, simplify=False)
            else:
                try:
                    NTcell.buildShape(bBox, surfTR=CTRF, simplify=False)
                except:
                    print(f"fail converting cell {NTcell.name}")
                    fails.append(NTcell.name)

            if NTcell.shape is None:
                continue

            if duplicate:
                if ContainerCell.CurrentTR:
                    cell = NTcell.copy()
                    cell.transformSolid(ContainerCell.CurrentTR)
                else:
                    cell = NTcell
            else:
                cell = NTcell

        if universeCut and ContainerCell.shape:
            cell.shape = interferencia(ContainerCell, cell)

        if not cell.FILL or ContainerCell.level + 1 == levelMax:
            CADUniverse.append(cell)
        else:
            if ContainerCell.CurrentTR:
                cell.CurrentTR = ContainerCell.CurrentTR.multiply(cell.TRFL)
            cell.level = ContainerCell.level + 1
            univ, ff = BuildUniverse((cell.FILL, levelMax), cell, AllUniverses, universeCut=universeCut)
            CADUniverse.append(univ)
            fails.extend(ff)

    return ((ContainerCell.name, Ustart), CADUniverse), fails


def makeTree(CADdoc, CADCells):

    label, universeCADCells = CADCells
    groupObj = CADdoc.addObject("App::Part", "Materials")

    groupObj.Label = f"Universe_{label[1]}_Container_{label[0]}"

    CADObj = {}
    for i, c in enumerate(universeCADCells):
        if type(c) is tuple:
            groupObj.addObject(makeTree(CADdoc, c))
        else:
            featObj = CADdoc.addObject("Part::FeaturePython", f"solid{i}")
            featObj.Label = f"Cell_{c.name}_{c.MAT}"
            featObj.Shape = c.shape
            if c.MAT not in CADObj.keys():
                CADObj[c.MAT] = [featObj]
            else:
                CADObj[c.MAT].append(featObj)

    for mat, matGroup in CADObj.items():
        groupMatObj = CADdoc.addObject("App::Part", "Materials")
        groupMatObj.Label = f"Material_{mat}"
        groupMatObj.addObjects(matGroup)
        groupObj.addObject(groupMatObj)

    return groupObj



#
# Set of useful functions used in different parts of the code
#
import logging
import math

import FreeCAD

from .basic_functions_part1 import (
    is_in_line,
    is_in_plane,
    is_in_tolerance,
    is_opposite,
    is_parallel,
    is_same_value,
)
from ..write.functions import mcnp_surface


def Fuzzy(index, dtype, surf1, surf2, val, tol, options, tolerances, numeric_format):

    fuzzy_logger = logging.getLogger("fuzzy_logger")

    same = val <= tol

    if dtype == "plane":
        p1str = mcnp_surface(index, "Plane", surf1, options, tolerances, numeric_format)
        p2str = mcnp_surface(0, "Plane", surf2, options, tolerances, numeric_format)
        fuzzy_logger.info(f"Same surface : {same}")
        fuzzy_logger.info(f"Plane distance / Tolerance : {val} {tol}\n {p1str}\n {p2str}\n\n")

    elif dtype == "cylRad":
        cyl1str = mcnp_surface(index, "Cylinder", surf1, options, tolerances, numeric_format)
        cyl2str = mcnp_surface(0, "Cylinder", surf2, options, tolerances, numeric_format)
        fuzzy_logger.info(f"Same surface : {same}")
        fuzzy_logger.info(f"Diff Radius / Tolerance: {val} {tol}")
        fuzzy_logger.info(f"{cyl1str}\n {cyl2str}\n\n")

    elif dtype == "cylAxs":
        cyl1str = mcnp_surface(index, "Cylinder", surf1, options, tolerances, numeric_format)
        cyl2str = mcnp_surface(0, "Cylinder", surf2, options, tolerances, numeric_format)
        fuzzy_logger.info(f"Same surface : {same}")
        fuzzy_logger.info(f"Dist Axis / Tolerance: {val} {tol}")
        fuzzy_logger.info(f"{cyl1str} {cyl2str}")


def is_same_plane(p1, p2, options, tolerances, numeric_format, fuzzy=(False, 0)):
    if is_parallel(p1.Axis, p2.Axis, tolerances.pln_angle):
        d1 = p1.Axis.dot(p1.Position)
        d2 = p2.Axis.dot(p2.Position)
        if is_opposite(p1.Axis, p2.Axis, tolerances.pln_angle):
            d2 = -d2
        d = abs(d1 - d2)
        if tolerances.relativeTol:
            tol = tolerances.pln_distance * max(p2.dimL1, p2.dimL2)
        else:
            tol = tolerances.pln_distance

        isSame, is_fuzzy = is_in_tolerance(d, tol, 0.5 * tol, 2 * tol)
        if is_fuzzy and fuzzy[0]:
            Fuzzy(fuzzy[1], "plane", p2, p1, d, tol, options, tolerances, numeric_format)
        return isSame
    return False


def is_same_cylinder(
    cyl1,
    cyl2,
    options,
    tolerances,
    numeric_format,
    fuzzy=(False, 0),
):
    if tolerances.relativeTol:
        rtol = tolerances.cyl_distance * max(cyl2.Radius, cyl1.Radius)
    else:
        rtol = tolerances.cyl_distance

    is_same_rad, is_fuzzy = is_in_tolerance(cyl2.Radius - cyl1.Radius, rtol, 0.5 * rtol, 2 * rtol)
    if is_fuzzy and fuzzy[0]:
        Fuzzy(
            fuzzy[1],
            "cylRad",
            cyl2,
            cyl1,
            abs(cyl2.Radius - cyl1.Radius),
            rtol,
            options,
            tolerances,
            numeric_format,
        )

    if is_same_rad:
        if is_parallel(cyl1.Axis, cyl2.Axis, tolerances.cyl_angle):
            c12 = cyl1.Center - cyl2.Center
            d = cyl1.Axis.cross(c12).Length

            if tolerances.relativeTol:
                tol = tolerances.cyl_distance * max(cyl1.Center.Length, cyl2.Center.Length)
            else:
                tol = tolerances.cyl_distance

            is_same_center, is_fuzzy = is_in_tolerance(d, tol, 0.5 * tol, 2 * tol)
            if is_fuzzy and fuzzy[0]:
                Fuzzy(
                    fuzzy[1],
                    "cylAxs",
                    cyl1,
                    cyl2,
                    d,
                    tol,
                    options,
                    tolerances,
                    numeric_format,
                )

            return is_same_center
    return False


def is_same_cone(cone1, cone2, dtol=1e-6, atol=1e-6, rel_tol=True):
    if is_same_value(cone1.SemiAngle, cone2.SemiAngle, atol):
        if is_parallel(cone1.Axis, cone2.Axis, atol):
            if rel_tol:
                tol = dtol * max(cone1.Apex.Length, cone2.Apex.Length)
            else:
                tol = dtol
            return cone1.Apex.isEqual(cone2.Apex, tol)
    return False


def is_same_sphere(sph1, sph2, tolerance=1e-6, rel_tol=True):
    if rel_tol:
        rtol = tolerance * max(sph2.Radius, sph1.Radius)
    else:
        rtol = tolerance
    if is_same_value(sph1.Radius, sph2.Radius, rtol):
        if rel_tol:
            ctol = tolerance * max(sph2.Center.Length, sph1.Center.Length)
        else:
            ctol = tolerance
        return sph1.Center.isEqual(sph2.Center, ctol)

    return False


def is_same_torus(tor1, tor2, dtol=1e-6, atol=1e-6, rel_tol=True):
    if is_parallel(tor1.Axis, tor2.Axis, atol):
        if tor1.Axis.dot(tor2.Axis) < 0:
            return False  # Assume same cone with oposite axis as different
        if rel_tol:
            Rtol = dtol * max(tor1.MajorRadius, tor2.MajorRadius)
            rtol = dtol * max(tor1.MinorRadius, tor2.MinorRadius)
        else:
            Rtol = dtol
            rtol = dtol

        if is_same_value(tor1.MajorRadius, tor2.MajorRadius, Rtol) and is_same_value(tor1.MinorRadius, tor2.MinorRadius, rtol):
            if rel_tol:
                ctol = dtol * max(tor1.Center.Length, tor2.Center.Length)
            else:
                ctol = dtol
            return tor1.Center.isEqual(tor2.Center, ctol)
    return False


def is_duplicate_in_list(num_str1, i, lista):
    for j, elem2 in enumerate(lista):
        if i == j:
            continue
        num_str2 = f"{elem2:11.4E}"
        num_str3 = f"{elem2 + 2.0 * math.pi:11.4E}"
        num_str4 = f"{elem2 - 2.0 * math.pi:11.4E}"

        if abs(float(num_str2)) < 1.0e-5:
            num_str2 = "%11.4E" % 0.0

        if abs(float(num_str3)) < 1.0e-5:
            num_str3 = "%11.4E" % 0.0

        if abs(float(num_str4)) < 1.0e-5:
            num_str4 = "%11.4E" % 0.0

        if num_str1 == num_str2 or num_str1 == num_str3 or num_str1 == num_str4:
            return True

    return False



import logging

import FreeCAD
import Part

from ..decompose import decom_one as Decom
from ..utils import basic_functions_part2 as BF
from ..utils import geometry_gu as GU
from ..utils.basic_functions_part1 import is_opposite, is_parallel
from ..utils.boolean_function import BoolSequence

logger = logging.getLogger("general_logger")


def commonEdge(face1, face2):
    for e1 in face1.Edges:
        for e2 in face2.Edges:
            if e1.isSame(e2):
                return e1
    return None


def isConvex(face1, face2, edge):
    de = 0.1
    tol = 1.0e-5
    delta_point = edge.Vertexes[1].Point - edge.Vertexes[0].Point
    delta_point.normalize()
    axis = delta_point.cross(face2.Surface.Axis)

    point = edge.CenterOfMass + de * axis
    if not face2.__face__.isInside(point, tol, True):
        axis = -axis

    convex = False
    if face1.Surface.Axis.dot(axis) < 0:
        if face1.Orientation == "Forward":
            convex = True
    else:
        if face1.Orientation == "Reversed":
            convex = True
    return convex


def removeElement(Faces, idf):
    for i, f in enumerate(Faces):
        if f[0] == idf:
            del Faces[i]
            break


def is_inverted(solid):
    face = solid.Faces[0]
    parameter_range = face.ParameterRange
    u = (parameter_range[1] + parameter_range[0]) / 2.0
    v = (parameter_range[3] + parameter_range[2]) / 2.0

    point2 = face.CenterOfMass.add(face.normalAt(u, v).multiply(1.0e-6))

    if solid.isInside(point2, 1e-7, False):
        return True
    else:
        return False


# TODO rename this function as there are two with the same name
def get_id(face_in, Surfaces, options, tolerances, numeric_format):

    if is_parallel(face_in.Axis, FreeCAD.Vector(1, 0, 0), tolerances.pln_angle):
        plane = "PX"
    elif is_parallel(face_in.Axis, FreeCAD.Vector(0, 1, 0), tolerances.pln_angle):
        plane = "PY"
    elif is_parallel(face_in.Axis, FreeCAD.Vector(0, 0, 1), tolerances.pln_angle):
        plane = "PZ"
    else:
        plane = "P"

    for s in Surfaces[plane]:
        if BF.is_same_plane(
            face_in,
            s.Surf,
            options=options,
            tolerances=tolerances,
            numeric_format=numeric_format,
        ):
            return s.Index

    return 0


def translate(meta_list, surfaces, universe_box, setting, options, tolerances, numeric_format):
    tot_solid = len(meta_list)
    for i, m in enumerate(meta_list):
        if m.IsEnclosure:
            continue
        logger.info(f"Decomposing solid: {i}/{tot_solid}")
        if setting.debug:
            logger.info(m.Comments)
            if m.IsEnclosure:
                m.Solids[0].exportStep(f"origEnclosure_{i}.stp")
            else:
                m.Solids[0].exportStep(f"origSolid_{i}.stp")

        surfaces.extend(
            Decom.extract_surfaces(
                Part.makeCompound(m.Solids),
                "Plane3Pts",
                universe_box,
                options,
                tolerances,
                numeric_format,
                MakeObj=False,
            )
        )
        set_definition(m, surfaces, options, tolerances)


# TODO rename this, but be careful as there are other functions in the code with the same name
def set_definition(meta_obj, surfaces, options, tolerances):
    solids = meta_obj.Solids
    s_def = BoolSequence(operator="OR")

    for sol in solids:
        subSol = BoolSequence(operator="AND")
        flag_inv = is_inverted(sol)
        solid_gu = GU.SolidGu(sol, tolerances=tolerances, plane3Pts=True)

        faces = []
        for face in solid_gu.Faces:
            if abs(face.Area) < 1e-2:
                continue
            if face.Area < 0:
                logger.warning("Negative surface Area")
            if str(face.Surface) != "<Plane object>":
                logger.warning("All surfaces must be plane")
                continue
            if face.Orientation not in ("Forward", "Reversed"):
                continue

            id = get_id(face.Surface, surfaces, options, tolerances)
            s = surfaces.get_surface(id)
            if is_opposite(face.Surface.Axis, s.Surf.Axis, tolerances.pln_angle):
                id = -id
            if face.Orientation == "Forward":
                id = -id
            if flag_inv:
                id = -id
            faces.append((id, face))

        while len(faces) > 0:
            id1, face1 = faces[0]
            no_convex = []
            for id2, face2 in faces[1:]:
                edge = commonEdge(face1, face2)
                if edge is None:
                    continue
                if not isConvex(face1, face2, edge):
                    no_convex.append(id2)

            if no_convex != []:
                no_convex.insert(0, id1)
                for i in no_convex:
                    removeElement(faces, i)
                orPlanes = BoolSequence(operator="OR")
                orPlanes.append(*no_convex)
                subSol.append(orPlanes)
            else:
                removeElement(faces, id1)
                subSol.append(id1)

        s_def.append(subSol)

    meta_obj.set_definition(s_def)



import re

mostinner = re.compile(r"\([^\(^\)]*\)")  # identify most inner parentheses
number = re.compile(r"(?P<value>[-+]?\d+)")
mix = re.compile(r"(?P<value>([-+]?\d+|\[0+\]))")
TFX = re.compile(r"(?P<value>[FTXo]+)")
PValue = re.compile(r"P\d+")
NValue = re.compile(r"N\d+")
conversion = {"T": True, "F": False, "X": None}


class BoolSequence:
    def __init__(self, definition=None, operator=None):
        if definition:
            self.elements = []
            self.set_def(definition)
        else:
            self.elements = []
            self.operator = operator
            self.level = 0

    def __str__(self):
        out = f"{self.operator}["
        if type(self.elements) is bool:
            return " True " if self.elements else " False "
        for e in self.elements:
            if type(e) is int or type(e) is bool or type(e) is str:
                out += f" {e} "
            else:
                out += e.__str__()

        out += "] "
        return out

    def append(self, *seq):
        for s in seq:
            if type(s) is int:
                level = -1
            else:
                level = s.level
                if type(s.elements) is bool:
                    if self.operator == "AND" and s.elements is False or self.operator == "OR" and s.elements is True:
                        self.level = 0
                        self.elements = s.elements
                        return
                    else:
                        continue

            self.elements.append(s)
            self.level = max(self.level, level + 1)

    def assign(self, Seq):
        self.operator = Seq.operator
        self.elements = Seq.elements
        self.level = Seq.level

    def copy(self):
        cp = BoolSequence()
        cp.operator = self.operator
        cp.level = self.level
        if type(self.elements) is bool:
            cp.elements = self.elements
        else:
            for e in self.elements:
                if type(e) is int:
                    cp.elements.append(e)
                else:
                    cp.elements.append(e.copy())
        return cp

    def get_complementary(self):

        c = BoolSequence(operator=self.comp_operator())
        c.level = self.level

        if self.level == 0:
            for e in self.elements:
                c.elements.append(-e)
            return c
        else:
            self.group_single()
            for e in self.elements:
                c.elements.append(e.get_complementary())
            return c

    def comp_operator(self):
        if self.operator == "AND":
            return "OR"
        else:
            return "AND"

    def simplify(self, CT, loop=0):
        surfNames = self.get_surfaces_numbers()
        if not surfNames:
            return
        #   print(CT)
        #    print('start',self)
        newNames = surfNames
        simplified = False
        for valname in surfNames:
            #      print('factorise',valname)
            if valname in newNames:
                chg = self.factorize(valname, CT)
                simplified = simplified or chg
                newNames = self.get_surfaces_numbers()
        self.join_operators()

        if self.level == 0 or (self.elements) is bool:
            return

        # new temp para ver como va
        # for e in self.elements:
        #   if type(e) is not int:
        #     e.simplify(CT)
        # return
        # end new

        if loop == 0:
            ANDSeq = BoolSequence(operator=self.operator)
            ORSeq = BoolSequence(operator=self.operator)
            for e in self.elements:
                if e.operator == "AND":
                    ANDSeq.append(e)
                else:
                    ORSeq.append(e)

            ANDSeq.simplify(CT, loop=1)
            ORSeq.simplify(CT, loop=1)
            newSeq = BoolSequence(operator=self.operator)
            if ANDSeq.elements:
                newSeq.append(ANDSeq)
            if ORSeq.elements:
                newSeq.append(ORSeq)
            newSeq.join_operators()
            self.assign(newSeq)
        else:
            for e in reversed(self.elements):
                e.simplify(CT, loop=0)
                if type(e.elements) is bool:
                    if self.operator == "AND" and e.elements is False or self.operator == "OR" and e.elements is True:
                        self.elements = e.elements
                        self.level = 0
                        break
                    else:
                        self.elements.remove(e)

            if self.elements == []:
                self.level = 0
                if self.operator == "AND":
                    self.elements = True
                else:
                    self.elements = False

    # check if level 0 sequence have oposite value a & -a = 0  , a|-a = 1
    # return the value of the sequence None(unknown), True, False
    def check(self):
        if self.level == 0:
            if type(self.elements) is bool:
                return self.elements

            signedSurf = set(self.elements)
            surfname = self.get_surfaces_numbers()
            if len(signedSurf) == len(surfname):
                return None  # means same surface has not positive and negative value
            elif self.operator == "AND":
                return False
            else:
                return True
        else:
            if type(self.elements) is bool:
                return self.elements

            self.group_single()
            noneVal = False
            for e in self.elements:
                res = e.check()
                if res is None:
                    noneVal = True
                elif self.operator == "AND" and res is False:
                    return False
                elif self.operator == "OR" and res is True:
                    return True

            if noneVal:
                return None
            elif self.operator == "AND":
                return True
            else:
                return False

    def substitute(self, var, val):
        if type(self.elements) is bool:
            return
        name = abs(var)
        ic = len(self.elements)
        for e in reversed(self.elements):
            ic -= 1
            if type(e) is int:
                if abs(e) == name:
                    if type(val) is int:
                        if name == e:
                            self.elements[ic] = val
                        else:
                            self.elements[ic] = -val

                    else:
                        if name == e:
                            boolValue = val
                        else:
                            boolValue = not val

                        if self.operator == "AND" and not boolValue:
                            self.elements = False
                            self.level = 0
                            return
                        elif self.operator == "OR" and boolValue:
                            self.elements = True
                            self.level = 0
                            return
                        else:
                            self.elements.remove(e)
            else:
                e.substitute(var, val)

        self.clean()
        self.level_update()

    def removeSurf(self, name):
        if type(self.elements) is bool:
            return
        ic = len(self.elements)
        for e in reversed(self.elements):
            ic -= 1
            if type(e) is int:
                if e == name:
                    if self.operator == "AND":
                        self.elements.remove(e)
                    elif self.operator == "OR":
                        self.elements = True
                        self.level = 0
                        return
            else:
                e.removeSurf(name)

        self.clean()
        self.level_update()

    # remove sequence whom elements are boolean values instead of list
    def clean(self):
        if type(self.elements) is bool:
            return self.elements
        for e in reversed(self.elements):
            if type(e) is int:
                continue
            eVal = e.clean()
            if type(eVal) is bool:
                if eVal and self.operator == "OR":
                    self.elements = True
                    self.level = 0
                    return True
                elif not eVal and self.operator == "AND":
                    self.elements = False
                    self.level = 0
                    return False
                self.elements.remove(e)

        if self.elements == []:
            if self.operator == "OR":
                self.elements = False
            else:
                self.elements = True
            self.level = 0
            return self.elements
        else:
            return None

    # join redundant operators in sequence
    def join_operators(self):
        if self.level == 0:
            return
        if type(self.elements) is bool:
            return
        self.group_single()
        ANDop = []
        ORop = []

        for e in self.elements:
            if e.operator == "AND":
                ANDop.append(e)
            else:
                ORop.append(e)

        if len(ANDop) > 1 and self.operator == "AND":
            newSeq = BoolSequence(operator="AND")
            for s in ANDop:
                newSeq.elements.extend(s.elements)
                self.elements.remove(s)
            newSeq.level_update()
            self.append(newSeq)

        elif len(ORop) > 1 and self.operator == "OR":
            newSeq = BoolSequence(operator="OR")
            for s in ORop:
                newSeq.elements.extend(s.elements)
                self.elements.remove(s)
            newSeq.level_update()
            self.append(newSeq)

        if self.level > 0 and len(self.elements) == 1:
            self.operator = self.elements[0].operator
            self.elements = self.elements[0].elements
            self.level -= 1
            self.join_operators()

        if self.level == 0:
            return
        for e in self.elements:
            e.join_operators()

    def factorize(self, valname, CT=None):

        if CT is None:
            trueSet = {abs(valname): True}
            falseSet = {abs(valname): False}
        else:
            trueSet, falseSet = CT.get_constraint_set(valname)

        if trueSet is None:
            self.substitute(valname, False)
            return True

        if falseSet is None:
            self.substitute(valname, True)
            return True

        funcVal = self.evaluate(trueSet, CT)
        if funcVal is False:
            newSeq = BoolSequence(operator="AND")
            # self.substitute(valname,False)
            for name, value in falseSet.items():
                self.substitute(name, value)
            newSeq.append(-valname, self.copy())
            self.assign(newSeq)
            return True
        elif funcVal is True:
            newSeq = BoolSequence(operator="OR")
            # self.substitute(valname,False)
            for name, value in falseSet.items():
                self.substitute(name, value)
            newSeq.append(valname, self.copy())
            self.assign(newSeq)
            return True

        funcVal = self.evaluate(falseSet, CT)
        if funcVal is False:
            newSeq = BoolSequence(operator="AND")
            # self.substitute(valname,True)
            for name, value in trueSet.items():
                self.substitute(name, value)
            newSeq.append(valname, self.copy())
            self.assign(newSeq)
            return True
        elif funcVal is True:
            newSeq = BoolSequence(operator="OR")
            # self.substitute(valname,True)
            for name, value in trueSet.items():
                self.substitute(name, value)
            newSeq.append(-valname, self.copy())
            self.assign(newSeq)
            return True

        return False

    def evaluate_newbad(self, valueSet, CT=None):

        if type(self.elements) is bool:
            return self.elements
        self.group_single()
        newSeq = self.copy()
        for name, value in valueSet.items():
            newSeq.substitute(name, value)

        if type(newSeq.elements) is bool:
            return newSeq.elements
        else:
            surfNames = newSeq.get_surfaces_numbers()
            valname = surfNames[0]
            if CT is None:
                trueSet = {abs(valname): True}
                falseSet = {abs(valname): False}
            else:
                trueSet, falseSet = CT.get_constraint_set(valname)

            if trueSet is None:
                trueVal = True
            else:
                trueVal = newSeq.evaluate(trueSet, CT)
                if trueVal is None:
                    return None

            if falseSet is None:
                falseVal = False
            else:
                falseVal = newSeq.evaluate(falseSet, CT)
                if falseVal is None:
                    return None

            if trueVal != falseVal:
                return None
            else:
                return trueVal

    def evaluate(self, valueSet, CT=None):
        if type(self.elements) is bool:
            return self.elements
        self.group_single()
        op = self.operator
        noneVal = False
        for e in self.elements:
            if type(e) is int:
                key = abs(e)
                if key in valueSet.keys():
                    val = valueSet[key]
                else:
                    val = None

                if val is not None:
                    if e < 0:
                        val = not val
                else:
                    noneVal = True
            else:
                val = e.evaluate(valueSet)

            if val is None:
                noneVal = True
            elif op == "AND" and val is False:
                return False
            elif op == "OR" and val is True:
                return True

        if noneVal:
            return None
        if op == "AND":
            return True
        else:
            return False

    def set_def(self, expression):
        terms, operator = outer_terms(expression)
        self.operator = operator
        self.level = 0
        for t in terms:
            if is_integer(t):
                self.elements.append(int(t.strip("(").strip(")")))
            else:
                x = BoolSequence(t)
                self.level = max(x.level + 1, self.level)
                self.elements.append(x)
        self.group_single()

    def group_single(self):
        if self.level == 0:
            return
        if type(self.elements) is bool:
            return
        group = []
        for e in reversed(self.elements):
            if type(e) is int:
                group.append(e)
                self.elements.remove(e)
            elif e.level == 0 and len(e.elements) == 1:
                group.append(e.elements[0])
                self.elements.remove(e)

        if not group:
            return
        seq = BoolSequence()
        seq.elements.extend(group)
        seq.operator = self.operator
        seq.level = 0
        self.elements.insert(0, seq)

    def sameOperator(self):
        if self.level == 0:
            return
        if type(self.elements) is bool:
            return

        changed = []
        new = []
        for i, e in enumerate(self.elements):
            if e.operator == self.operator:
                changed.append(i)
                for s in e.elements:
                    if type(s) is int:
                        new.append(BoolSequence(str(s)))
                    else:
                        new.append(s)

        for i in reversed(changed):
            del self.elements[i]
        self.append(*new)
        self.level_update()

        return

    def get_surfaces_numbers(self):
        if type(self.elements) is bool:
            return tuple()
        surf = set()
        for e in self.elements:
            if type(e) is int:
                surf.add(abs(e))
            else:
                surf.update(e.get_surfaces_numbers())
        return tuple(surf)

    def level_update(self):
        if type(self.elements) is bool:
            self.level = 0
            return

        newlev = 0
        for e in self.elements:
            if type(e) is int:
                lev = 0
            else:
                lev = e.level + 1
            newlev = max(lev, newlev)
        self.level = newlev


def outer_terms(expression, value="number"):
    if value == "number":
        # reValue = number
        reValue = mix
        nullVal = "0"
    else:
        reValue = TFX
        nullVal = "o"

    expr = expression

    # Loop until no redundant parentheses are found
    cont = True

    while cont:
        # Loop over most inner parentheses
        pos = 0
        cont = False
        while True:
            m = mostinner.search(expr, pos)
            if not m:
                break
            cont = True
            if redundant(m, expr):
                # remove redundant parentheses
                expr = expr[: m.start()] + " " + expr[m.start() + 1 : m.end() - 1] + " " + expr[m.end() :]
            else:
                # replace no redundant parentheses by 0 and : by ;
                zeros = "[" + nullVal * (m.end() - m.start() - 2) + "]"
                expr = expr[: m.start()] + zeros + expr[m.end() :]

            pos = m.end()

    if ":" in expr:
        terms = []
        pos = 0
        while True:
            newpos = expr.find(":", pos)
            if newpos == -1:
                terms.append(expression[pos:].strip())
                break
            terms.append(expression[pos:newpos].strip())
            pos = newpos + 1
        return (terms, "OR")
    else:
        terms = []
        pos = 0
        while True:
            m = reValue.search(expr, pos)
            if not m:
                break
            terms.append(expression[m.start() : m.end()])
            pos = m.end()
        return (terms, "AND")


def redundant(m, geom):
    """check if the inner parentheses are redundant"""
    term = m.group()

    # Find first valid character at the left of the  parenthese
    leftOK = True
    left = m.start() - 1
    while left > -1:
        if geom[left] in ("\n", "C", "$", " "):
            left -= 1
        else:
            if geom[left] not in ("(", ":"):
                leftOK = False
            break

    # check if no ':' (or) are inside the parenthese
    # if not, parentheses are redundants
    if term.find(":") == -1:
        return True

    # Find first valid character at the right of the  parenthese
    rightOK = True
    right = m.end()
    while right < len(geom):
        if geom[right] in ("\n", "C", "$", " "):
            right += 1
        else:
            if geom[right] not in (")", ":"):
                rightOK = False
            break

    # if parentheses are like:
    # {( or : } ( ....... ) {) or :}
    # parentheses are redundants

    if leftOK and rightOK:
        return True
    else:
        return False


def is_integer(x):
    try:
        int(x.strip("(").strip(")"))
        return True
    except:
        return False



#
# Set of useful functions used in different parts of the code
#
import logging
import math

import BOPTools.SplitAPI
import FreeCAD
import numpy as np
import Part

logger = logging.getLogger("general_logger")

from .basic_functions_part1 import (
    ConeParams,
    CylinderParams,
    Plane3PtsParams,
    PlaneParams,
    SphereParams,
    TorusParams,
    is_parallel,
)
from . import basic_functions_part2 as BF


def get_box(comp, enlargeBox):
    bb = FreeCAD.BoundBox(comp.BoundBox)
    bb.enlarge(enlargeBox)
    xMin, yMin, zMin = bb.XMin, bb.YMin, bb.ZMin
    xLength, yLength, zLength = bb.XLength, bb.YLength, bb.ZLength

    return Part.makeBox(
        xLength,
        yLength,
        zLength,
        FreeCAD.Vector(xMin, yMin, zMin),
        FreeCAD.Vector(0, 0, 1),
    )


class GeounedSolid:

    def __init__(self, id, comsolid=None):
        refine = True
        if not comsolid:
            self.Solids = None
            self.Volume = None
            self.BoundBox = None
        elif type(comsolid) is list:
            self.Solids = comsolid
            vol = 0
            self.BoundBox = FreeCAD.BoundBox()
            for s in comsolid:
                vol += s.Volume
                self.BoundBox.add(s.BoundBox)
            self.Volume = vol
        else:
            if refine:
                try:
                    self.Solids = comsolid.removeSplitter().Solids
                except:
                    self.Solids = comsolid.Solids

                for s in self.Solids:
                    if s.Volume < 0:
                        s.reverse()
            else:
                self.Solids = comsolid.Solids
            self.Volume = comsolid.Volume
            self.BoundBox = comsolid.BoundBox

        self.__id__ = id
        self.label = None
        self.Definition = []
        self.Faces = []
        self.Comments = ""
        self.Density = 0
        self.Dilution = 1
        self.Material = 0
        self.Surfaces = []
        self.Rho = None
        self.MatInfo = None
        self.CellType = "solid"  # other types : 'void', 'enclosure', 'fill'
        self.Universe = 0
        self.Void = False
        self.IsEnclosure = False
        self.EnclosureID = None
        self.ParentEnclosureID = None
        self.SonEnclosures = []
        self.enclosure_list = None
        self.CADSolid = None
        self.UniverseBox = None
        self.NullCell = True

    def update_solids(self, solidList):
        self.Solids = solidList
        vol = 0
        self.BoundBox = FreeCAD.BoundBox()
        for s in solidList:
            vol += s.Volume
            self.BoundBox.add(s.BoundBox)

    def set_cad_solid(self):
        self.CADSolid = Part.makeCompound(self.Solids)
        self.Volume = self.CADSolid.Volume
        self.BoundBox = self.CADSolid.BoundBox

    def optimalBoundingBox(self):
        if self.CADSolid is None:
            self.set_cad_solid()
        return self.CADSolid.optimalBoundingBox()

    def set_definition(self, definition, simplify=False):

        if definition is None:
            self.NullCell = True
            return

        self.NullCell = False
        self.Definition = definition

        if not self.Void:
            if not self.Definition.elements:
                self.NullCell = True
                return

        self.Surfaces = tuple(self.Definition.get_surfaces_numbers())

    def set_faces(self, faces):
        self.Faces = faces

    def set_comments(self, comments):
        self.Comments = comments

    def set_material(self, material, rho=None, info=None):

        self.Material = material
        self.Rho = rho
        self.MatInfo = info

        if rho is not None:
            self.Density = self.Dilution * rho
        else:
            if material != 0:
                self.Density = None

    def set_dilution(self, dilution):
        self.Dilution = dilution
        if self.Rho is not None:
            self.Density = self.Rho * dilution

    def check_intersection(self, solid, dtolerance=1.0e-6, vtolerance=1e-10):
        """Check if solid intersect with current solid.
        return : -2 solid fully embedded in self.CADSolid ;
                 -1 self.CADSolid fully embedded in solid ;
                  0 self.CADSolid intersect solid ;
                  1 self.CADSolid and solid fully disjoint"""

        dist = 1e12
        for sol1 in self.CADSolid.Solids:
            for sol2 in solid.Solids:
                try:
                    distShape = sol1.distToShape(sol2)[0]
                except:
                    logger.info("Failed solid1.distToshape(solid2), try with inverted solids")
                    distShape = sol2.distToShape(sol1)[0]
                    logger.info(f"inverted disToShape OK {distShape}")
                dist = min(dist, distShape)
                if dist == 0:
                    break

        if dist > dtolerance:
            return 1

        common = self.CADSolid.common(solid)
        if abs(common.Volume) < vtolerance:
            return 1
        if abs(self.CADSolid.Volume - common.Volume) / common.Volume < vtolerance:
            return -1
        elif abs(solid.Volume - common.Volume) / common.Volume < vtolerance:
            return -2
        else:
            return 0


class GeounedSurface:

    def __init__(self, params, boundBox, Face=None):

        self.Index = 0
        self.__boundBox__ = boundBox
        if params[0] == "Plane":
            self.Type = "Plane"
            self.Surf = PlaneParams(params[1])  # plane point defined as the shortest distance to origin
        elif params[0] == "Plane3Pts":
            self.Type = "Plane"
            self.Surf = Plane3PtsParams(params[1])  # plane point defined with 3 points
        elif params[0] == "Cylinder":
            self.Type = params[0]
            self.Surf = CylinderParams(params[1])
        elif params[0] == "Cone":
            self.Type = params[0]
            self.Surf = ConeParams(params[1])
        elif params[0] == "Sphere":
            self.Type = params[0]
            self.Surf = SphereParams(params[1])
        elif params[0] == "Torus":
            self.Type = params[0]
            self.Surf = TorusParams(params[1])

        self.shape = Face

        if type(Face) is str:
            if Face == "Build":
                self.build_surface()

        if self.shape == "Build":
            raise ValueError(f"stop {params} {boundBox}")
        return

    def build_surface(self):

        if self.Type == "Plane":

            normal = self.Surf.Axis
            p0 = normal.dot(self.Surf.Position)
            Box = FreeCAD.BoundBox(self.__boundBox__)
            Box.enlarge(10)

            pointEdge = []
            for i in range(12):
                edge = Box.getEdge(i)
                p1 = normal.dot(edge[0])
                p2 = normal.dot(edge[1])
                d0 = p0 - p1
                d1 = p2 - p1
                if d1 != 0:
                    a = d0 / d1
                    if a >= 0 and a <= 1:
                        pointEdge.append(edge[0] + a * (edge[1] - edge[0]))

            if len(pointEdge) == 0:
                self.shape = None  # Plane does not cross box
                return

            s = FreeCAD.Vector((0, 0, 0))
            for v in pointEdge:
                s = s + v
            s = s / len(pointEdge)

            vtxvec = []
            for v in pointEdge:
                vtxvec.append(v - s)

            X0 = vtxvec[0]
            Y0 = normal.cross(X0)

            orden = []
            for i, v in enumerate(vtxvec):
                phi = np.arctan2(v.dot(Y0), v.dot(X0))
                orden.append((phi, i))
            orden.sort()

            self.shape = Part.Face(Part.makePolygon([pointEdge[p[1]] for p in orden], True))

            return

        elif self.Type == "Cylinder":
            dir = self.Surf.Axis
            pnt = self.Surf.Center
            rad = self.Surf.Radius

            dmin = dir.dot((self.__boundBox__.getPoint(0) - pnt))
            dmax = dmin
            for i in range(1, 8):
                d = dir.dot((self.__boundBox__.getPoint(i) - pnt))
                dmin = min(d, dmin)
                dmax = max(d, dmax)

            pnt = pnt + (dmin - 5) * dir
            length = dmax - dmin + 10
            self.shape = Part.makeCylinder(rad, length, pnt, dir, 360.0)
            return

        elif self.Type == "Cone":
            #         dir  = self.Surf.Axis
            #         apex = self.Surf.Apex
            #         tan = math.tan(self.Surf.SemiAngle)
            #         rad = tan*diag # New radius far away to cover all geometry

            axis = self.Surf.Axis
            apex = self.Surf.Apex
            tan = math.tan(self.Surf.SemiAngle)

            dmax = axis.dot((self.__boundBox__.getPoint(0) - apex))
            for i in range(1, 8):
                d = axis.dot((self.__boundBox__.getPoint(i) - apex))
                dmax = max(d, dmax)

            if dmax > 0:
                length = dmax + 10
                rad = tan * length
                self.shape = Part.makeCone(0.0, rad, length, apex, axis, 360.0)
            else:
                self.shape = None
            return

        elif self.Type == "Sphere":
            rad = self.Surf.Radius
            pnt = self.Surf.Center
            self.shape = Part.makeSphere(rad, pnt).Faces[0]
            return

        elif self.Type == "Torus":
            axis = self.Surf.Axis
            center = self.Surf.Center
            majorR = self.Surf.MajorRadius
            minorR = self.Surf.MinorRadius

            torus = Part.makeTorus(majorR, minorR, center, axis)
            self.shape = torus.Faces[0]
            return
        else:
            logger.error(f"Type {self.Type} is not defined")
            return


class SurfacesDict(dict):
    def __init__(self, surfaces=None, offset=0):
        self.IndexOffset = offset
        surfname = ["PX", "PY", "PZ", "P", "Cyl", "Cone", "Sph", "Tor"]
        for name in surfname:
            self[name] = []

        self.__surfIndex__ = dict()

        if surfaces is not None:
            for key in surfaces.keys():
                self[key] = surfaces[key][:]
                self.__surfIndex__[key] = surfaces.__surfIndex__[key][:]
            self.surfaceNumber = surfaces.surfaceNumber
            self.__last_obj__ = (surfaces.__last_obj__[0], surfaces.__last_obj__[1])
        else:
            self.surfaceNumber = 0
            self.__last_obj__ = ("", -1)
            for key in surfname:
                self.__surfIndex__[key] = []
        return

    def __str__(self):
        for key in self.keys():
            logger.info(f"{key}, {self[key]}")
        return ""

    def get_surface(self, index):

        lastKey = self.__last_obj__[0]
        lastInd = self.__last_obj__[1]
        if lastKey != "":
            if len(self[lastKey]) > 0:
                if self[lastKey][lastInd].Index == index:
                    return self[lastKey][lastInd]

        for key, values in self.__surfIndex__.items():
            if index not in values:
                continue
            i = values.index(index)
            self.__last_obj__ = (key, i)
            return self[key][i]

        logger.info(f"Index {index} not found in Surfaces")
        return None

    def del_surface(self, index):
        self.get_surface(index)
        self.__surfIndex__[self.__last_obj__[0]].remove(index)
        del self[self.__last_obj__[0]][self.__last_obj__[1]]
        return

    def extend(self, surface, options, tolerances, numeric_format):
        for Pkey in ["PX", "PY", "PZ", "P"]:
            for s in surface[Pkey]:
                self.add_plane(s, options, tolerances, numeric_format, False)
        for s in surface["Cyl"]:
            self.add_cylinder(s, options, tolerances, numeric_format, False)
        for s in surface["Cone"]:
            self.add_cone(s, tolerances)
        for s in surface["Sph"]:
            self.add_sphere(s, tolerances)
        for s in surface["Tor"]:
            self.add_torus(s, tolerances)

    def add_plane(self, plane, options, tolerances, numeric_format, fuzzy):
        ex = FreeCAD.Vector(1, 0, 0)
        ey = FreeCAD.Vector(0, 1, 0)
        ez = FreeCAD.Vector(0, 0, 1)

        if is_parallel(plane.Surf.Axis, ex, tolerances.pln_angle):
            add_plane = True
            for i, p in enumerate(self["PX"]):
                if BF.is_same_plane(
                    plane.Surf,
                    p.Surf,
                    options=options,
                    tolerances=tolerances,
                    numeric_format=numeric_format,
                    fuzzy=(fuzzy, p.Index),
                ):
                    add_plane = False
                    index = p.Index
                    self.__last_obj__ = ("PX", i)
                    break
            if add_plane:
                self.surfaceNumber += 1
                plane.Index = self.surfaceNumber + self.IndexOffset
                self.__last_obj__ = ("PX", len(self["PX"]))
                self["PX"].append(plane)
                self.__surfIndex__["PX"].append(plane.Index)

        elif is_parallel(plane.Surf.Axis, ey, tolerances.pln_angle):
            add_plane = True
            for i, p in enumerate(self["PY"]):
                if BF.is_same_plane(
                    plane.Surf,
                    p.Surf,
                    options=options,
                    tolerances=tolerances,
                    numeric_format=numeric_format,
                    fuzzy=(fuzzy, p.Index),
                ):
                    add_plane = False
                    index = p.Index
                    self.__last_obj__ = ("PY", i)
                    break
            if add_plane:
                self.surfaceNumber += 1
                plane.Index = self.surfaceNumber + self.IndexOffset
                self.__last_obj__ = ("PY", len(self["PY"]))
                self["PY"].append(plane)
                self.__surfIndex__["PY"].append(plane.Index)

        elif is_parallel(plane.Surf.Axis, ez, tolerances.pln_angle):
            add_plane = True
            for i, p in enumerate(self["PZ"]):
                if BF.is_same_plane(
                    plane.Surf,
                    p.Surf,
                    options=options,
                    tolerances=tolerances,
                    numeric_format=numeric_format,
                    fuzzy=(fuzzy, p.Index),
                ):
                    add_plane = False
                    index = p.Index
                    self.__last_obj__ = ("PZ", i)
                    break
            if add_plane:
                self.surfaceNumber += 1
                plane.Index = self.surfaceNumber + self.IndexOffset
                self.__last_obj__ = ("PZ", len(self["PZ"]))
                self["PZ"].append(plane)
                self.__surfIndex__["PZ"].append(plane.Index)

        else:
            add_plane = True
            for i, p in enumerate(self["P"]):
                if BF.is_same_plane(
                    plane.Surf,
                    p.Surf,
                    options=options,
                    tolerances=tolerances,
                    numeric_format=numeric_format,
                    fuzzy=(fuzzy, p.Index),
                ):
                    add_plane = False
                    index = p.Index
                    self.__last_obj__ = ("P", i)
                    break
            if add_plane:
                self.surfaceNumber += 1
                plane.Index = self.surfaceNumber + self.IndexOffset
                self.__last_obj__ = ("P", len(self["P"]))
                self["P"].append(plane)
                self.__surfIndex__["P"].append(plane.Index)

        if add_plane:
            return plane.Index, False
        else:
            return index, True

    def add_cylinder(self, cyl, options, tolerances, numeric_format, fuzzy=False):
        addCyl = True
        for i, c in enumerate(self["Cyl"]):
            if BF.is_same_cylinder(
                cyl.Surf,
                c.Surf,
                options=options,
                tolerances=tolerances,
                numeric_format=numeric_format,
                fuzzy=(fuzzy, c.Index),
            ):
                addCyl = False
                index = c.Index
                self.__last_obj__ = ("Cyl", i)
                break

        if addCyl:
            self.surfaceNumber += 1
            cyl.Index = self.surfaceNumber + self.IndexOffset
            self.__last_obj__ = ("Cyl", len(self["Cyl"]))
            self["Cyl"].append(cyl)
            self.__surfIndex__["Cyl"].append(cyl.Index)
            return cyl.Index, False
        else:
            return index, True

    def add_cone(self, cone, tolerances):
        cone_added = True
        for i, c in enumerate(self["Cone"]):
            if BF.is_same_cone(
                cone.Surf,
                c.Surf,
                dtol=tolerances.kne_distance,
                atol=tolerances.kne_angle,
                rel_tol=tolerances.relativeTol,
            ):
                cone_added = False
                index = c.Index
                self.__last_obj__ = ("Cone", i)
                break
        if cone_added:
            self.surfaceNumber += 1
            cone.Index = self.surfaceNumber + self.IndexOffset
            self.__last_obj__ = ("Cone", len(self["Cone"]))
            self["Cone"].append(cone)
            self.__surfIndex__["Cone"].append(cone.Index)
            return cone.Index, False
        else:
            return index, True

    def add_sphere(self, sph, tolerances):
        sphere_added = True
        for i, s in enumerate(self["Sph"]):
            if BF.is_same_sphere(
                sph.Surf,
                s.Surf,
                tolerances.sph_distance,
                rel_tol=tolerances.relativeTol,
            ):
                sphere_added = False
                index = s.Index
                self.__last_obj__ = ("Sph", i)
                break
        if sphere_added:
            self.surfaceNumber += 1
            sph.Index = self.surfaceNumber + self.IndexOffset
            self.__last_obj__ = ("Sph", len(self["Sph"]))
            self["Sph"].append(sph)
            self.__surfIndex__["Sph"].append(sph.Index)
            return sph.Index, False
        else:
            return index, True

    def add_torus(self, tor, tolerances):
        add_torus = True
        for i, s in enumerate(self["Tor"]):
            if BF.is_same_torus(
                tor.Surf,
                s.Surf,
                dtol=tolerances.tor_distance,
                atol=tolerances.tor_angle,
                rel_tol=tolerances.relativeTol,
            ):
                add_torus = False
                index = s.Index
                self.__last_obj__ = ("Tor", i)
                break
        if add_torus:
            self.surfaceNumber += 1
            tor.Index = self.surfaceNumber + self.IndexOffset
            self.__last_obj__ = ("Tor", len(self["Tor"]))
            self["Tor"].append(tor)
            self.__surfIndex__["Tor"].append(tor.Index)
            return tor.Index, False
        else:
            return index, True


def split_bop(solid, tools, tolerance, options, scale=0.1):

    if tolerance >= 0.1:
        compSolid = BOPTools.SplitAPI.slice(solid, tools, "Split", tolerance=tolerance)

    elif tolerance < 1e-12:
        if options.scaleUp:
            tol = 1e-13 if options.splitTolerance == 0 else options.splitTolerance
            compSolid = split_bop(solid, tools, tol / scale, options, 1.0 / scale)
        else:
            compSolid = BOPTools.SplitAPI.slice(solid, tools, "Split", tolerance=tolerance)

    else:
        try:
            compSolid = BOPTools.SplitAPI.slice(solid, tools, "Split", tolerance=tolerance)
        except:
            compSolid = split_bop(solid, tools, tolerance * scale, options, scale)

    return compSolid
